@@ -16,20 +16,17 @@ package main
 import (
 	"bufio"
 	"bytes"
+	"context"
 	"errors"
 	"fmt"
-	"github.com/alecthomas/kingpin"
 	"io"
 	"log"
 	"net"
-	"net/http"
-	"os"
 	"regexp"
 	"strconv"
 	"strings"
 	"time"
 
-	"github.com/hpcloud/tail"
 	"github.com/prometheus/client_golang/prometheus"
 )
 
@@ -43,13 +40,14 @@ var (
 // PostfixExporter holds the state that should be preserved by the
 // Postfix Prometheus metrics exporter across scrapes.
 type PostfixExporter struct {
-	showqPath string
-	journal   *Journal
-	tailer    *tail.Tail
+	showqPath           string
+	logSrc              LogSource
+	logUnsupportedLines bool
 
 	// Metrics that should persist after refreshes, based on logs.
 	cleanupProcesses                prometheus.Counter
 	cleanupRejects                  prometheus.Counter
+	cleanupNotAccepted              prometheus.Counter
 	lmtpDelays                      *prometheus.HistogramVec
 	pipeDelays                      *prometheus.HistogramVec
 	qmgrInsertsNrcpt                prometheus.Histogram
@@ -57,6 +55,8 @@ type PostfixExporter struct {
 	qmgrRemoves                     prometheus.Counter
 	smtpDelays                      *prometheus.HistogramVec
 	smtpTLSConnects                 *prometheus.CounterVec
+	smtpConnectionTimedOut          prometheus.Counter
+	smtpDeferreds                   prometheus.Counter
 	smtpdConnects                   prometheus.Counter
 	smtpdDisconnects                prometheus.Counter
 	smtpdFCrDNSErrors               prometheus.Counter
@@ -66,6 +66,18 @@ type PostfixExporter struct {
 	smtpdSASLAuthenticationFailures prometheus.Counter
 	smtpdTLSConnects                *prometheus.CounterVec
 	unsupportedLogEntries           *prometheus.CounterVec
+	smtpStatusDeferred              prometheus.Counter
+	opendkimSignatureAdded          *prometheus.CounterVec
+}
+
+// A LogSource is an interface to read log lines.
+type LogSource interface {
+	// Path returns a representation of the log location.
+	Path() string
+
+	// Read returns the next log line. Returns `io.EOF` at the end of
+	// the log.
+	Read(context.Context) (string, error)
 }
 
 // CollectShowqFromReader parses the output of Postfix's 'showq' command
@@ -90,12 +102,6 @@ func CollectShowqFromReader(file io.Reader, ch chan<- prometheus.Metric) error {
 
 // CollectTextualShowqFromReader parses Postfix's textual showq output.
 func CollectTextualShowqFromReader(file io.Reader, ch chan<- prometheus.Metric) error {
-	scanner := bufio.NewScanner(file)
-	scanner.Split(bufio.ScanLines)
-
-	// Regular expression for matching postqueue's output. Example:
-	// "A07A81514      5156 Tue Feb 14 13:13:54  MAILER-DAEMON"
-	messageLine := regexp.MustCompile(`^[0-9A-F]+([\*!]?) +(\d+) (\w{3} \w{3} +\d+ +\d+:\d{2}:\d{2}) +`)
 
 	// Histograms tracking the messages by size and age.
 	sizeHistogram := prometheus.NewHistogramVec(
@@ -115,56 +121,72 @@ func CollectTextualShowqFromReader(file io.Reader, ch chan<- prometheus.Metric)
 		},
 		[]string{"queue"})
 
+	err := CollectTextualShowqFromScanner(sizeHistogram, ageHistogram, file)
+
+	sizeHistogram.Collect(ch)
+	ageHistogram.Collect(ch)
+	return err
+}
+
+func CollectTextualShowqFromScanner(sizeHistogram prometheus.ObserverVec, ageHistogram prometheus.ObserverVec, file io.Reader) error {
+	scanner := bufio.NewScanner(file)
+	scanner.Split(bufio.ScanLines)
 	// Initialize all queue buckets to zero.
 	for _, q := range []string{"active", "hold", "other"} {
 		sizeHistogram.WithLabelValues(q)
 		ageHistogram.WithLabelValues(q)
 	}
 
-	now := time.Now()
 	location, err := time.LoadLocation("Local")
 	if err != nil {
 		log.Println(err)
 	}
 
+	// Regular expression for matching postqueue's output. Example:
+	// "A07A81514      5156 Tue Feb 14 13:13:54  MAILER-DAEMON"
+	messageLine := regexp.MustCompile(`^[0-9A-F]+([\*!]?) +(\d+) (\w{3} \w{3} +\d+ +\d+:\d{2}:\d{2}) +`)
+
 	for scanner.Scan() {
-		matches := messageLine.FindStringSubmatch(scanner.Text())
-		if matches != nil {
-			// Derive the name of the message queue.
-			queue := "other"
-			if matches[1] == "*" {
-				queue = "active"
-			} else if matches[1] == "!" {
-				queue = "hold"
-			}
-
-			// Parse the message size.
-			size, err := strconv.ParseFloat(matches[2], 64)
-			if err != nil {
-				return err
-			}
-
-			// Parse the message date. Unfortunately, the
-			// output contains no year number. Assume it
-			// applies to the last year for which the
-			// message date doesn't exceed time.Now().
-			date, err := time.ParseInLocation("Mon Jan 2 15:04:05",
-				matches[3], location)
-			if err != nil {
-				return err
-			}
-			date = date.AddDate(now.Year(), 0, 0)
-			if date.After(now) {
-				date = date.AddDate(-1, 0, 0)
-			}
-
-			sizeHistogram.WithLabelValues(queue).Observe(size)
-			ageHistogram.WithLabelValues(queue).Observe(now.Sub(date).Seconds())
+		text := scanner.Text()
+		matches := messageLine.FindStringSubmatch(text)
+		if matches == nil {
+			continue
 		}
-	}
+		queueMatch := matches[1]
+		sizeMatch := matches[2]
+		dateMatch := matches[3]
 
-	sizeHistogram.Collect(ch)
-	ageHistogram.Collect(ch)
+		// Derive the name of the message queue.
+		queue := "other"
+		if queueMatch == "*" {
+			queue = "active"
+		} else if queueMatch == "!" {
+			queue = "hold"
+		}
+
+		// Parse the message size.
+		size, err := strconv.ParseFloat(sizeMatch, 64)
+		if err != nil {
+			return err
+		}
+
+		// Parse the message date. Unfortunately, the
+		// output contains no year number. Assume it
+		// applies to the last year for which the
+		// message date doesn't exceed time.Now().
+		date, err := time.ParseInLocation("Mon Jan 2 15:04:05", dateMatch, location)
+		if err != nil {
+			return err
+		}
+		now := time.Now()
+		date = date.AddDate(now.Year(), 0, 0)
+		if date.After(now) {
+			date = date.AddDate(-1, 0, 0)
+		}
+
+		sizeHistogram.WithLabelValues(queue).Observe(size)
+		ageHistogram.WithLabelValues(queue).Observe(now.Sub(date).Seconds())
+	}
 	return scanner.Err()
 }
 
@@ -252,16 +274,6 @@ func CollectBinaryShowqFromReader(file io.Reader, ch chan<- prometheus.Metric) e
 	return scanner.Err()
 }
 
-// CollectShowqFromFile collects Postfix queue statistics from a file.
-//func CollectShowqFromFile(path string, ch chan<- prometheus.Metric) error {
-//	fd, err := os.Open(path)
-//	if err != nil {
-//		return err
-//	}
-//	defer fd.Close()
-//	return CollectShowqFromReader(fd, ch)
-//}
-
 // CollectShowqFromSocket collects Postfix queue statistics from a socket.
 func CollectShowqFromSocket(path string, ch chan<- prometheus.Metric) error {
 	fd, err := net.Dial("unix", path)
@@ -274,187 +286,155 @@ func CollectShowqFromSocket(path string, ch chan<- prometheus.Metric) error {
 
 // Patterns for parsing log messages.
 var (
-	logLine                             = regexp.MustCompile(` ?postfix/(\w+)\[\d+\]: (.*)`)
+	logLine                             = regexp.MustCompile(` ?(postfix|opendkim)(/(\w+))?\[\d+\]: (.*)`)
 	lmtpPipeSMTPLine                    = regexp.MustCompile(`, relay=(\S+), .*, delays=([0-9\.]+)/([0-9\.]+)/([0-9\.]+)/([0-9\.]+), `)
 	qmgrInsertLine                      = regexp.MustCompile(`:.*, size=(\d+), nrcpt=(\d+) `)
-	smtpTLSLine                         = regexp.MustCompile(`^(\S+) TLS connection established to \S+: (\S+) with cipher (\S+) \((\d+)/(\d+) bits\)$`)
+	smtpStatusDeferredLine              = regexp.MustCompile(`, status=deferred`)
+	smtpTLSLine                         = regexp.MustCompile(`^(\S+) TLS connection established to \S+: (\S+) with cipher (\S+) \((\d+)/(\d+) bits\)`)
+	smtpConnectionTimedOut              = regexp.MustCompile(`^connect\s+to\s+(.*)\[(.*)\]:(\d+):\s+(Connection timed out)$`)
 	smtpdFCrDNSErrorsLine               = regexp.MustCompile(`^warning: hostname \S+ does not resolve to address `)
-	smtpdProcessesSASLLine              = regexp.MustCompile(`: client=.*, sasl_username=(\S+)`)
+	smtpdProcessesSASLLine              = regexp.MustCompile(`: client=.*, sasl_method=(\S+)`)
 	smtpdRejectsLine                    = regexp.MustCompile(`^NOQUEUE: reject: RCPT from \S+: ([0-9]+) `)
 	smtpdLostConnectionLine             = regexp.MustCompile(`^lost connection after (\w+) from `)
 	smtpdSASLAuthenticationFailuresLine = regexp.MustCompile(`^warning: \S+: SASL \S+ authentication failed: `)
-	smtpdTLSLine                        = regexp.MustCompile(`^(\S+) TLS connection established from \S+: (\S+) with cipher (\S+) \((\d+)/(\d+) bits\)$`)
+	smtpdTLSLine                        = regexp.MustCompile(`^(\S+) TLS connection established from \S+: (\S+) with cipher (\S+) \((\d+)/(\d+) bits\)`)
+	opendkimSignatureAdded              = regexp.MustCompile(`^[\w\d]+: DKIM-Signature field added \(s=(\w+), d=(.*)\)$`)
 )
 
 // CollectFromLogline collects metrict from a Postfix log line.
-func (e *PostfixExporter) CollectFromLogline(line string) {
+func (e *PostfixExporter) CollectFromLogLine(line string) {
 	// Strip off timestamp, hostname, etc.
-	if logMatches := logLine.FindStringSubmatch(line); logMatches != nil {
+	logMatches := logLine.FindStringSubmatch(line)
+
+	if logMatches == nil {
+		// Unknown log entry format.
+		e.addToUnsupportedLine(line, "")
+		return
+	}
+	process := logMatches[1]
+	remainder := logMatches[4]
+	switch process {
+	case "postfix":
 		// Group patterns to check by Postfix service.
-		if logMatches[1] == "cleanup" {
-			if strings.Contains(logMatches[2], ": message-id=<") {
+		subprocess := logMatches[3]
+		switch subprocess {
+		case "cleanup":
+			if strings.Contains(remainder, ": message-id=<") {
 				e.cleanupProcesses.Inc()
-			} else if strings.Contains(logMatches[2], ": reject: ") {
+			} else if strings.Contains(remainder, ": reject: ") {
 				e.cleanupRejects.Inc()
 			} else {
-				e.unsupportedLogEntries.WithLabelValues(logMatches[1]).Inc()
+				e.addToUnsupportedLine(line, subprocess)
 			}
-		} else if logMatches[1] == "lmtp" {
-			if lmtpMatches := lmtpPipeSMTPLine.FindStringSubmatch(logMatches[2]); lmtpMatches != nil {
-				pdelay, err := strconv.ParseFloat(lmtpMatches[2], 64)
-				if err != nil {
-					log.Printf("Couldn't convert LMTP pdelay: %v", err)
-				}
-				e.lmtpDelays.WithLabelValues("before_queue_manager").Observe(pdelay)
-				adelay, err := strconv.ParseFloat(lmtpMatches[3], 64)
-				if err != nil {
-					log.Printf("Couldn't convert LMTP adelay: %v", err)
-				}
-				e.lmtpDelays.WithLabelValues("queue_manager").Observe(adelay)
-				sdelay, err := strconv.ParseFloat(lmtpMatches[4], 64)
-				if err != nil {
-					log.Printf("Couldn't convert LMTP adelay: %v", err)
-				}
-				e.lmtpDelays.WithLabelValues("connection_setup").Observe(sdelay)
-				xdelay, err := strconv.ParseFloat(lmtpMatches[5], 64)
-				if err != nil {
-					log.Printf("Couldn't convert LMTP xdelay: %v", err)
-				}
-				e.lmtpDelays.WithLabelValues("transmission").Observe(xdelay)
+		case "lmtp":
+			if lmtpMatches := lmtpPipeSMTPLine.FindStringSubmatch(remainder); lmtpMatches != nil {
+				addToHistogramVec(e.lmtpDelays, lmtpMatches[2], "LMTP pdelay", "before_queue_manager")
+				addToHistogramVec(e.lmtpDelays, lmtpMatches[3], "LMTP adelay", "queue_manager")
+				addToHistogramVec(e.lmtpDelays, lmtpMatches[4], "LMTP sdelay", "connection_setup")
+				addToHistogramVec(e.lmtpDelays, lmtpMatches[5], "LMTP xdelay", "transmission")
 			} else {
-				e.unsupportedLogEntries.WithLabelValues(logMatches[1]).Inc()
+				e.addToUnsupportedLine(line, subprocess)
 			}
-		} else if logMatches[1] == "pipe" {
-			if pipeMatches := lmtpPipeSMTPLine.FindStringSubmatch(logMatches[2]); pipeMatches != nil {
-				pdelay, err := strconv.ParseFloat(pipeMatches[2], 64)
-				if err != nil {
-					log.Printf("Couldn't convert PIPE pdelay: %v", err)
-				}
-				e.pipeDelays.WithLabelValues(pipeMatches[1], "before_queue_manager").Observe(pdelay)
-				adelay, err := strconv.ParseFloat(pipeMatches[3], 64)
-				if err != nil {
-					log.Printf("Couldn't convert PIPE adelay: %v", err)
-				}
-				e.pipeDelays.WithLabelValues(pipeMatches[1], "queue_manager").Observe(adelay)
-				sdelay, err := strconv.ParseFloat(pipeMatches[4], 64)
-				if err != nil {
-					log.Printf("Couldn't convert PIPE sdelay: %v", err)
-				}
-				e.pipeDelays.WithLabelValues(pipeMatches[1], "connection_setup").Observe(sdelay)
-				xdelay, err := strconv.ParseFloat(pipeMatches[5], 64)
-				if err != nil {
-					log.Printf("Couldn't convert PIPE xdelay: %v", err)
-				}
-				e.pipeDelays.WithLabelValues(pipeMatches[1], "transmission").Observe(xdelay)
+		case "pipe":
+			if pipeMatches := lmtpPipeSMTPLine.FindStringSubmatch(remainder); pipeMatches != nil {
+				addToHistogramVec(e.pipeDelays, pipeMatches[2], "PIPE pdelay", pipeMatches[1], "before_queue_manager")
+				addToHistogramVec(e.pipeDelays, pipeMatches[3], "PIPE adelay", pipeMatches[1], "queue_manager")
+				addToHistogramVec(e.pipeDelays, pipeMatches[4], "PIPE sdelay", pipeMatches[1], "connection_setup")
+				addToHistogramVec(e.pipeDelays, pipeMatches[5], "PIPE xdelay", pipeMatches[1], "transmission")
 			} else {
-				e.unsupportedLogEntries.WithLabelValues(logMatches[1]).Inc()
+				e.addToUnsupportedLine(line, subprocess)
 			}
-		} else if logMatches[1] == "qmgr" {
-			if qmgrInsertMatches := qmgrInsertLine.FindStringSubmatch(logMatches[2]); qmgrInsertMatches != nil {
-				size, err := strconv.ParseFloat(qmgrInsertMatches[1], 64)
-				if err != nil {
-					log.Printf("Couldn't convert QMGR size: %v", err)
-				}
-				e.qmgrInsertsSize.Observe(size)
-				nrcpt, err := strconv.ParseFloat(qmgrInsertMatches[2], 64)
-				if err != nil {
-					log.Printf("Couldn't convert QMGR nrcpt: %v", err)
-				}
-				e.qmgrInsertsNrcpt.Observe(nrcpt)
-			} else if strings.HasSuffix(logMatches[2], ": removed") {
+		case "qmgr":
+			if qmgrInsertMatches := qmgrInsertLine.FindStringSubmatch(remainder); qmgrInsertMatches != nil {
+				addToHistogram(e.qmgrInsertsSize, qmgrInsertMatches[1], "QMGR size")
+				addToHistogram(e.qmgrInsertsNrcpt, qmgrInsertMatches[2], "QMGR nrcpt")
+			} else if strings.HasSuffix(remainder, ": removed") {
 				e.qmgrRemoves.Inc()
 			} else {
-				e.unsupportedLogEntries.WithLabelValues(logMatches[1]).Inc()
+				e.addToUnsupportedLine(line, subprocess)
 			}
-		} else if logMatches[1] == "smtp" {
-			if smtpMatches := lmtpPipeSMTPLine.FindStringSubmatch(logMatches[2]); smtpMatches != nil {
-				pdelay, err := strconv.ParseFloat(smtpMatches[2], 64)
-				if err != nil {
-					log.Printf("Couldn't convert SMTP pdelay: %v", err)
+		case "smtp":
+			if smtpMatches := lmtpPipeSMTPLine.FindStringSubmatch(remainder); smtpMatches != nil {
+				addToHistogramVec(e.smtpDelays, smtpMatches[2], "before_queue_manager", "")
+				addToHistogramVec(e.smtpDelays, smtpMatches[3], "queue_manager", "")
+				addToHistogramVec(e.smtpDelays, smtpMatches[4], "connection_setup", "")
+				addToHistogramVec(e.smtpDelays, smtpMatches[5], "transmission", "")
+				if smtpMatches := smtpStatusDeferredLine.FindStringSubmatch(remainder); smtpMatches != nil {
+					e.smtpStatusDeferred.Inc()
 				}
-				e.smtpDelays.WithLabelValues("before_queue_manager").Observe(pdelay)
-				adelay, err := strconv.ParseFloat(smtpMatches[3], 64)
-				if err != nil {
-					log.Printf("Couldn't convert SMTP adelay: %v", err)
-				}
-				e.smtpDelays.WithLabelValues("queue_manager").Observe(adelay)
-				sdelay, err := strconv.ParseFloat(smtpMatches[4], 64)
-				if err != nil {
-					log.Printf("Couldn't convert SMTP sdelay: %v", err)
-				}
-				e.smtpDelays.WithLabelValues("connection_setup").Observe(sdelay)
-				xdelay, err := strconv.ParseFloat(smtpMatches[5], 64)
-				if err != nil {
-					log.Printf("Couldn't convert SMTP xdelay: %v", err)
-				}
-				e.smtpDelays.WithLabelValues("transmission").Observe(xdelay)
-			} else if smtpTLSMatches := smtpTLSLine.FindStringSubmatch(logMatches[2]); smtpTLSMatches != nil {
+			} else if smtpTLSMatches := smtpTLSLine.FindStringSubmatch(remainder); smtpTLSMatches != nil {
 				e.smtpTLSConnects.WithLabelValues(smtpTLSMatches[1:]...).Inc()
+			} else if smtpMatches := smtpConnectionTimedOut.FindStringSubmatch(remainder); smtpMatches != nil {
+				e.smtpConnectionTimedOut.Inc()
 			} else {
-				e.unsupportedLogEntries.WithLabelValues(logMatches[1]).Inc()
+				e.addToUnsupportedLine(line, subprocess)
 			}
-		} else if logMatches[1] == "smtpd" {
-			if strings.HasPrefix(logMatches[2], "connect from ") {
+		case "smtpd":
+			if strings.HasPrefix(remainder, "connect from ") {
 				e.smtpdConnects.Inc()
-			} else if strings.HasPrefix(logMatches[2], "disconnect from ") {
+			} else if strings.HasPrefix(remainder, "disconnect from ") {
 				e.smtpdDisconnects.Inc()
-			} else if smtpdFCrDNSErrorsLine.MatchString(logMatches[2]) {
+			} else if smtpdFCrDNSErrorsLine.MatchString(remainder) {
 				e.smtpdFCrDNSErrors.Inc()
-			} else if smtpdLostConnectionMatches := smtpdLostConnectionLine.FindStringSubmatch(logMatches[2]); smtpdLostConnectionMatches != nil {
+			} else if smtpdLostConnectionMatches := smtpdLostConnectionLine.FindStringSubmatch(remainder); smtpdLostConnectionMatches != nil {
 				e.smtpdLostConnections.WithLabelValues(smtpdLostConnectionMatches[1]).Inc()
-			} else if smtpdProcessesSASLMatches := smtpdProcessesSASLLine.FindStringSubmatch(logMatches[2]); smtpdProcessesSASLMatches != nil {
+			} else if smtpdProcessesSASLMatches := smtpdProcessesSASLLine.FindStringSubmatch(remainder); smtpdProcessesSASLMatches != nil {
 				e.smtpdProcesses.WithLabelValues(smtpdProcessesSASLMatches[1]).Inc()
-			} else if strings.Contains(logMatches[2], ": client=") {
+			} else if strings.Contains(remainder, ": client=") {
 				e.smtpdProcesses.WithLabelValues("").Inc()
-			} else if smtpdRejectsMatches := smtpdRejectsLine.FindStringSubmatch(logMatches[2]); smtpdRejectsMatches != nil {
+			} else if smtpdRejectsMatches := smtpdRejectsLine.FindStringSubmatch(remainder); smtpdRejectsMatches != nil {
 				e.smtpdRejects.WithLabelValues(smtpdRejectsMatches[1]).Inc()
-			} else if smtpdSASLAuthenticationFailuresLine.MatchString(logMatches[2]) {
+			} else if smtpdSASLAuthenticationFailuresLine.MatchString(remainder) {
 				e.smtpdSASLAuthenticationFailures.Inc()
-			} else if smtpdTLSMatches := smtpdTLSLine.FindStringSubmatch(logMatches[2]); smtpdTLSMatches != nil {
+			} else if smtpdTLSMatches := smtpdTLSLine.FindStringSubmatch(remainder); smtpdTLSMatches != nil {
 				e.smtpdTLSConnects.WithLabelValues(smtpdTLSMatches[1:]...).Inc()
 			} else {
-				e.unsupportedLogEntries.WithLabelValues(logMatches[1]).Inc()
+				e.addToUnsupportedLine(line, subprocess)
 			}
-		} else {
-			// Unknown Postfix service.
-			e.unsupportedLogEntries.WithLabelValues(logMatches[1]).Inc()
+		default:
+			e.addToUnsupportedLine(line, subprocess)
 		}
-	} else {
+	case "opendkim":
+		if opendkimMatches := opendkimSignatureAdded.FindStringSubmatch(remainder); opendkimMatches != nil {
+			e.opendkimSignatureAdded.WithLabelValues(opendkimMatches[1], opendkimMatches[2]).Inc()
+		} else {
+			e.addToUnsupportedLine(line, process)
+		}
+	default:
 		// Unknown log entry format.
-		e.unsupportedLogEntries.WithLabelValues("").Inc()
+		e.addToUnsupportedLine(line, "")
 	}
 }
 
-// CollectLogfileFromFile tails a Postfix log file and collects entries from it.
-func (e *PostfixExporter) CollectLogfileFromFile() error {
-	for {
-		select {
-		case line := <-e.tailer.Lines:
-			e.CollectFromLogline(line.Text)
-		default:
-			return nil
-		}
+func (e *PostfixExporter) addToUnsupportedLine(line string, subprocess string) {
+	if e.logUnsupportedLines {
+		log.Printf("Unsupported Line: %v", line)
 	}
+	e.unsupportedLogEntries.WithLabelValues(subprocess).Inc()
+}
+
+func addToHistogram(h prometheus.Histogram, value, fieldName string) {
+	float, err := strconv.ParseFloat(value, 64)
+	if err != nil {
+		log.Printf("Couldn't convert value '%s' for %v: %v", value, fieldName, err)
+	}
+	h.Observe(float)
+}
+func addToHistogramVec(h *prometheus.HistogramVec, value, fieldName string, labels ...string) {
+	float, err := strconv.ParseFloat(value, 64)
+	if err != nil {
+		log.Printf("Couldn't convert value '%s' for %v: %v", value, fieldName, err)
+	}
+	h.WithLabelValues(labels...).Observe(float)
 }
 
 // NewPostfixExporter creates a new Postfix exporter instance.
-func NewPostfixExporter(showqPath string, logfilePath string, journal *Journal) (*PostfixExporter, error) {
-	var tailer *tail.Tail
-	if logfilePath != "" {
-		var err error
-		tailer, err = tail.TailFile(logfilePath, tail.Config{
-			ReOpen:    true, // reopen the file if it's rotated
-			MustExist: true, // fail immediately if the file is missing or has incorrect permissions
-			Follow:    true, // run in follow mode
-		})
-		if err != nil {
-			return nil, err
-		}
-	}
+func NewPostfixExporter(showqPath string, logSrc LogSource, logUnsupportedLines bool) (*PostfixExporter, error) {
+	timeBuckets := []float64{1e-3, 1e-2, 1e-1, 1.0, 10, 1 * 60, 1 * 60 * 60, 24 * 60 * 60, 2 * 24 * 60 * 60}
 	return &PostfixExporter{
-		showqPath: showqPath,
-		tailer:    tailer,
-		journal:   journal,
+		logUnsupportedLines: logUnsupportedLines,
+		showqPath:           showqPath,
+		logSrc:              logSrc,
 
 		cleanupProcesses: prometheus.NewCounter(prometheus.CounterOpts{
 			Namespace: "postfix",
@@ -466,12 +446,17 @@ func NewPostfixExporter(showqPath string, logfilePath string, journal *Journal)
 			Name:      "cleanup_messages_rejected_total",
 			Help:      "Total number of messages rejected by cleanup.",
 		}),
+		cleanupNotAccepted: prometheus.NewCounter(prometheus.CounterOpts{
+			Namespace: "postfix",
+			Name:      "cleanup_messages_not_accepted_total",
+			Help:      "Total number of messages not accepted by cleanup.",
+		}),
 		lmtpDelays: prometheus.NewHistogramVec(
 			prometheus.HistogramOpts{
 				Namespace: "postfix",
 				Name:      "lmtp_delivery_delay_seconds",
 				Help:      "LMTP message processing time in seconds.",
-				Buckets:   []float64{1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3},
+				Buckets:   timeBuckets,
 			},
 			[]string{"stage"}),
 		pipeDelays: prometheus.NewHistogramVec(
@@ -479,7 +464,7 @@ func NewPostfixExporter(showqPath string, logfilePath string, journal *Journal)
 				Namespace: "postfix",
 				Name:      "pipe_delivery_delay_seconds",
 				Help:      "Pipe message processing time in seconds.",
-				Buckets:   []float64{1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3},
+				Buckets:   timeBuckets,
 			},
 			[]string{"relay", "stage"}),
 		qmgrInsertsNrcpt: prometheus.NewHistogram(prometheus.HistogramOpts{
@@ -504,7 +489,7 @@ func NewPostfixExporter(showqPath string, logfilePath string, journal *Journal)
 				Namespace: "postfix",
 				Name:      "smtp_delivery_delay_seconds",
 				Help:      "SMTP message processing time in seconds.",
-				Buckets:   []float64{1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3},
+				Buckets:   timeBuckets,
 			},
 			[]string{"stage"}),
 		smtpTLSConnects: prometheus.NewCounterVec(
@@ -514,6 +499,16 @@ func NewPostfixExporter(showqPath string, logfilePath string, journal *Journal)
 				Help:      "Total number of outgoing TLS connections.",
 			},
 			[]string{"trust", "protocol", "cipher", "secret_bits", "algorithm_bits"}),
+		smtpDeferreds: prometheus.NewCounter(prometheus.CounterOpts{
+			Namespace: "postfix",
+			Name:      "smtp_deferred_messages_total",
+			Help:      "Total number of messages that have been deferred on SMTP.",
+		}),
+		smtpConnectionTimedOut: prometheus.NewCounter(prometheus.CounterOpts{
+			Namespace: "postfix",
+			Name:      "smtp_connection_timed_out_total",
+			Help:      "Total number of messages that have been deferred on SMTP.",
+		}),
 		smtpdConnects: prometheus.NewCounter(prometheus.CounterOpts{
 			Namespace: "postfix",
 			Name:      "smtpd_connects_total",
@@ -542,7 +537,7 @@ func NewPostfixExporter(showqPath string, logfilePath string, journal *Journal)
 				Name:      "smtpd_messages_processed_total",
 				Help:      "Total number of messages processed.",
 			},
-			[]string{"sasl_username"}),
+			[]string{"sasl_method"}),
 		smtpdRejects: prometheus.NewCounterVec(
 			prometheus.CounterOpts{
 				Namespace: "postfix",
@@ -569,6 +564,19 @@ func NewPostfixExporter(showqPath string, logfilePath string, journal *Journal)
 				Help:      "Log entries that could not be processed.",
 			},
 			[]string{"service"}),
+		smtpStatusDeferred: prometheus.NewCounter(prometheus.CounterOpts{
+			Namespace: "postfix",
+			Name:      "smtp_status_deferred",
+			Help:      "Total number of messages deferred.",
+		}),
+		opendkimSignatureAdded: prometheus.NewCounterVec(
+			prometheus.CounterOpts{
+				Namespace: "opendkim",
+				Name:      "signatures_added_total",
+				Help:      "Total number of messages signed.",
+			},
+			[]string{"subject", "domain"},
+		),
 	}, nil
 }
 
@@ -576,8 +584,12 @@ func NewPostfixExporter(showqPath string, logfilePath string, journal *Journal)
 func (e *PostfixExporter) Describe(ch chan<- *prometheus.Desc) {
 	ch <- postfixUpDesc
 
+	if e.logSrc == nil {
+		return
+	}
 	ch <- e.cleanupProcesses.Desc()
 	ch <- e.cleanupRejects.Desc()
+	ch <- e.cleanupNotAccepted.Desc()
 	e.lmtpDelays.Describe(ch)
 	e.pipeDelays.Describe(ch)
 	ch <- e.qmgrInsertsNrcpt.Desc()
@@ -585,6 +597,7 @@ func (e *PostfixExporter) Describe(ch chan<- *prometheus.Desc) {
 	ch <- e.qmgrRemoves.Desc()
 	e.smtpDelays.Describe(ch)
 	e.smtpTLSConnects.Describe(ch)
+	ch <- e.smtpDeferreds.Desc()
 	ch <- e.smtpdConnects.Desc()
 	ch <- e.smtpdDisconnects.Desc()
 	ch <- e.smtpdFCrDNSErrors.Desc()
@@ -593,7 +606,39 @@ func (e *PostfixExporter) Describe(ch chan<- *prometheus.Desc) {
 	e.smtpdRejects.Describe(ch)
 	ch <- e.smtpdSASLAuthenticationFailures.Desc()
 	e.smtpdTLSConnects.Describe(ch)
+	ch <- e.smtpStatusDeferred.Desc()
 	e.unsupportedLogEntries.Describe(ch)
+	e.smtpConnectionTimedOut.Describe(ch)
+	e.opendkimSignatureAdded.Describe(ch)
+}
+
+func (e *PostfixExporter) StartMetricCollection(ctx context.Context) {
+	if e.logSrc == nil {
+		return
+	}
+
+	gaugeVec := prometheus.NewGaugeVec(
+		prometheus.GaugeOpts{
+			Namespace: "postfix",
+			Subsystem: "",
+			Name:      "up",
+			Help:      "Whether scraping Postfix's metrics was successful.",
+		},
+		[]string{"path"})
+	gauge := gaugeVec.WithLabelValues(e.logSrc.Path())
+	defer gauge.Set(0)
+
+	for {
+		line, err := e.logSrc.Read(ctx)
+		if err != nil {
+			if err != io.EOF {
+				log.Printf("Couldn't read journal: %v", err)
+			}
+			return
+		}
+		e.CollectFromLogLine(line)
+		gauge.Set(1)
+	}
 }
 
 // Collect metrics from Postfix's showq socket and its log file.
@@ -614,31 +659,12 @@ func (e *PostfixExporter) Collect(ch chan<- prometheus.Metric) {
 			e.showqPath)
 	}
 
-	var src string
-	if e.journal != nil {
-		err = e.CollectLogfileFromJournal()
-		src = e.journal.Path
-	} else {
-		err = e.CollectLogfileFromFile()
-		src = e.tailer.Filename
+	if e.logSrc == nil {
+		return
 	}
-	if err == nil {
-		ch <- prometheus.MustNewConstMetric(
-			postfixUpDesc,
-			prometheus.GaugeValue,
-			1.0,
-			src)
-	} else {
-		log.Printf("Failed to scrape log: %s", err)
-		ch <- prometheus.MustNewConstMetric(
-			postfixUpDesc,
-			prometheus.GaugeValue,
-			0.0,
-			src)
-	}
-
 	ch <- e.cleanupProcesses
 	ch <- e.cleanupRejects
+	ch <- e.cleanupNotAccepted
 	e.lmtpDelays.Collect(ch)
 	e.pipeDelays.Collect(ch)
 	ch <- e.qmgrInsertsNrcpt
@@ -646,6 +672,7 @@ func (e *PostfixExporter) Collect(ch chan<- prometheus.Metric) {
 	ch <- e.qmgrRemoves
 	e.smtpDelays.Collect(ch)
 	e.smtpTLSConnects.Collect(ch)
+	ch <- e.smtpDeferreds
 	ch <- e.smtpdConnects
 	ch <- e.smtpdDisconnects
 	ch <- e.smtpdFCrDNSErrors
@@ -654,58 +681,8 @@ func (e *PostfixExporter) Collect(ch chan<- prometheus.Metric) {
 	e.smtpdRejects.Collect(ch)
 	ch <- e.smtpdSASLAuthenticationFailures
 	e.smtpdTLSConnects.Collect(ch)
+	ch <- e.smtpStatusDeferred
 	e.unsupportedLogEntries.Collect(ch)
-}
-
-func main() {
-	var (
-		app                                           = kingpin.New("postfix_exporter", "Prometheus metrics exporter for postfix")
-		listenAddress                                 = app.Flag("web.listen-address", "Address to listen on for web interface and telemetry.").Default(":9154").String()
-		metricsPath                                   = app.Flag("web.telemetry-path", "Path under which to expose metrics.").Default("/metrics").String()
-		postfixShowqPath                              = app.Flag("postfix.showq_path", "Path at which Postfix places its showq socket.").Default("/var/spool/postfix/public/showq").String()
-		postfixLogfilePath                            = app.Flag("postfix.logfile_path", "Path where Postfix writes log entries. This file will be truncated by this exporter.").Default("/var/log/postfix_exporter_input.log").String()
-		systemdEnable                                 bool
-		systemdUnit, systemdSlice, systemdJournalPath string
-	)
-	systemdFlags(&systemdEnable, &systemdUnit, &systemdSlice, &systemdJournalPath, app)
-
-	kingpin.MustParse(app.Parse(os.Args[1:]))
-
-	var journal *Journal
-	if systemdEnable {
-		var err error
-		journal, err = NewJournal(systemdUnit, systemdSlice, systemdJournalPath)
-		if err != nil {
-			log.Fatalf("Error opening systemd journal: %s", err)
-		}
-		defer journal.Close()
-	}
-
-	exporter, err := NewPostfixExporter(
-		*postfixShowqPath,
-		*postfixLogfilePath,
-		journal,
-	)
-	if err != nil {
-		log.Fatalf("Failed to create PostfixExporter: %s", err)
-	}
-	prometheus.MustRegister(exporter)
-
-	http.Handle(*metricsPath, prometheus.Handler())
-	http.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
-		_, err = w.Write([]byte(`
-			<html>
-			<head><title>Postfix Exporter</title></head>
-			<body>
-			<h1>Postfix Exporter</h1>
-			<p><a href='` + *metricsPath + `'>Metrics</a></p>
-			</body>
-			</html>`))
-		if err != nil {
-			panic(err)
-		}
-	})
-
-	log.Print("Listening on ", *listenAddress)
-	log.Fatal(http.ListenAndServe(*listenAddress, nil))
+	ch <- e.smtpConnectionTimedOut
+	e.opendkimSignatureAdded.Collect(ch)
 }
